import { React, type AllWidgetProps } from 'jimu-core'
import { useEffect, useRef, useState } from 'react';
import { JimuMapViewComponent, JimuMapView } from 'jimu-arcgis'

import MapView from "@arcgis/core/views/MapView";
import Point from "@arcgis/core/geometry/Point";
import Graphic from '@arcgis/core/Graphic';
import GraphicsLayer from '@arcgis/core/layers/GraphicsLayer';
import Extent from '@arcgis/core/geometry/Extent';
import SpatialReference from '@arcgis/core/geometry/SpatialReference';

import { canProject,project } from '@arcgis/core/geometry/support/webMercatorUtils';
import { watch } from '@arcgis/core/core/reactiveUtils';

import { MapObjectUtils } from './MapObjectUtils';

type Props = {
	widgetProps: AllWidgetProps<any>,
	orbitgt
}

/*
 * Copyright (c) 2018-2024, Orbit Geospatial Technologies. All rights reserved.
 * Use is subject to license terms.
 */

/**
 * WidgetView implements loading the Orbit Viewer and integrating with the Esri Map.
 */
const WidgetView = ({
	widgetProps,
	orbitgt
}: Props) => {

	/** Reference to the element the viewer is inserted into. */
	const elementRef = useRef<HTMLDivElement>();
	/** Reference to the top-level element for the widget. */
	const widgetRef = useRef<HTMLDivElement>();

	const [width,setWidth] = useState(null);
	const [height,setHeight] = useState(null);

	const [mapView,setMapView] = useState<MapView>(null);
	const [viewer,setViewer] = useState(null);

	const [mapObjectUtils] = useState(MapObjectUtils(orbitgt));
	const [waitingForPickPosition,setWaitingForPickPosition] = useState(false);
	const [mapObjectLayers] = useState(new Map<string,any>());
	const updatingMap = useRef(false);

	//
    // Get references to the 3DMapping Viewer SDK Types we will be using
	//
    var SDKViewer = orbitgt.mapping3d.sdk.viewer.SDKViewer;
    var Bounds = orbitgt.mapping3d.sdk.viewer.Bounds;
    var Coordinate = orbitgt.mapping3d.sdk.viewer.Coordinate;
    var MapObject = orbitgt.mapping3d.sdk.viewer.MapObject;
    var AList = orbitgt.mapping3d.sdk.viewer.AList;
    var AMap = orbitgt.mapping3d.sdk.viewer.AMap;
    var Constants = orbitgt.mapping3d.sdk.viewer.Constants;

	/**
	 * Update viewer size on size changes.
	 */
	useEffect(() => {
		if (viewer && width!=null && height!=null) {
			viewer.setSize(width,height);
		}
	},[viewer,width,height]);

	/**
	 * Create the viewer.
	 */
	useEffect(() => {
		if (elementRef!=null) {
			// Prepare startup options.
			var options = new AMap();
			options.set(Constants.STARTUP_APP_STATE,window.localStorage.getItem("orbit_mapping3d_state"));
			options.set(Constants.STARTUP_CRS,"4326");
			options.set(Constants.STARTUP_THROTTLE_ANIMATION_SIGNALS,true);

			// Choose an application name that best suits your plugin
			var applicationName = "Orbit 3D Mapping Experience Builder Widget";
			// Create viewer instance
			const viewer = new SDKViewer(applicationName,elementRef.current,options);
			viewer.isReady.then(() => setViewer(viewer));

			() => {
				return viewer.close();
			}
		}
	}, [elementRef]);

	/**
	 * Add Orbit Viewer signal handlers.
	 */
	useEffect(() => {
		if (viewer) {
			
			console.log("Viewer is ready!",viewer);

			const handlers = new Array<{remove: ()=>void}>();
			const addSignalHandler = (signal, handler): {remove: ()=>void} => {
				signal.add(handler);
				return {
					remove: () => signal.remove(handler)
				}
			};

			handlers.push(addSignalHandler(viewer.onAppStateChanged,saveState));
			handlers.push(addSignalHandler(viewer.onWorkspaceOpened,handleWorkspaceOpened));

			// Listen to interaction signals
			handlers.push(addSignalHandler(viewer.onZoomToExtent,handleZoomToExtent));
			handlers.push(addSignalHandler(viewer.onPanToPosition,handlePanToPosition));
			handlers.push(addSignalHandler(viewer.onPickPositionRequested,handlePickPositionRequested));
			handlers.push(addSignalHandler(viewer.onPickPositionCancelled,handlePickPositionCancelled));

			// Listen to overlay signals
			handlers.push(addSignalHandler(viewer.onFootprintsChanged,updateFootprints));
			handlers.push(addSignalHandler(viewer.onGlobalsChanged,updateGlobals));
			handlers.push(addSignalHandler(viewer.onMeasureStateChanged,updateMeasurements));
			handlers.push(addSignalHandler(viewer.onFieldOfViewsChanged,updateViews));

			return () => {
				handlers.forEach(handler => handler.remove());
			}
		}
	},[viewer,mapView]); // update when orbit or esri map is updated

	/**
	 * Add Esri feature layers.
	 */
	useEffect(() => {
		if (mapView) {
			// Add layers
			findLayer(MapObject.TYPE_FOOTPRINT);
			findLayer(MapObject.TYPE_MEASUREMENTS);
			findLayer(MapObject.TYPE_GLOBAL);
			findLayer(MapObject.TYPE_FOV);
			return () => {
				// Remove layers
				for(const layer of mapObjectLayers.values()) mapView.map.layers.remove(layer);
			}
		}
	},[mapView]);

	/**
	 * Add Esri map event handlers.
	 */
	useEffect(() => {
		if (mapView) {
			// Listen to map reposition
			let reactiveHandles: Array<{remove: ()=>void}> = new Array();
			reactiveHandles.push(watch(() => mapView.stationary, (stationary) => stationary && onMapMove()));
			// Listen to map click
			reactiveHandles.push(mapView.on("click",onMapClick));
			return () => {
				// Remove event handlers
				reactiveHandles.forEach(handle => handle.remove());
			}
		}
	}); // we re-attach these event handlers on every render, so the callbacks always have access to the current state.

	/**
	 * Update the viewer size when the widget container changes size.
	 */
	useEffect(() => {
		if (widgetRef) {
			const resizeObserver = new ResizeObserver(() => {
				if (widgetRef.current) {
					setWidth(widgetRef.current.offsetWidth);
					setHeight(widgetRef.current.offsetHeight);
				}
			});
			resizeObserver.observe(widgetRef.current);
			return () => {
				resizeObserver.disconnect();
			}
		}
	},[widgetRef]);

	/**
	 * Handle a "pick position on the map" signal from the viewer.
	 */
	const handlePickPositionRequested = () => {
		mapView.container.style.cursor = "crosshair";
		setWaitingForPickPosition(true);
	}

	/**
	 * Handle a "pick position on the map" signal from the viewer.
	 */
	const handlePickPositionCancelled = () => {
		mapView.container.style.cursor = "auto";
		setWaitingForPickPosition(false);
	}

	const handleWorkspaceOpened = () => {
		console.log("Workspace is open.");
	}

	/** 
	 * Handle a "application state changed" signal from the viewer.
	 */
	const saveState = (state) => {
		// save it.
		window.localStorage.setItem("orbit_mapping3d_state",state);
	}
	
	/** 
	 * This function returns a layer for a specific MapObject type.
	 * @param mapObjectType the MapObject type.
	 * @return returns an esri Layer instance, already added to the map.
	 */
	const findLayer = (mapObjectType) => {
		// The layer does not exist?
		if (!mapObjectLayers.has(mapObjectType)) {
			// Create a new layer
			var layer = new GraphicsLayer();
			// Save a reference
			mapObjectLayers.set(mapObjectType,layer);
			// Add to map
			mapView.map.layers.add(layer);
		}
		// Return the layer
		return mapObjectLayers.get(mapObjectType);
	}

	/** 
	 * Generic function that adds a list of MapObject instances to a map layer.
	 * @param {GraphicLayer} layer a layer for Graphic objects.
	 * @param {LabelLayer} labelLayer a layer for labels.
	 * @param {AList} mapObjects the list of map objects.
	 */
	const addMapObjectsToLayer = (layer,labelLayer,mapObjects) => {

		// Add objects
		for (var i=0; i<mapObjects.size(); i++)
		{
			var mapObject = mapObjects.get(i);
			var graphic = mapObjectUtils.getGraphicForMapObject(mapObject);
			var labelGraphic = null;

			if (graphic) {
				// The MapObject has a label and we have a layer for labels?
				if (mapObject.label && labelLayer) {
					// Create a label graphic. 
					labelGraphic = new Graphic();
					// Attach the label to the graphic's position (in case of a point) 
					// or else to the graphic extent's center.
					if (graphic.geometry instanceof Point) labelGraphic.geometry = graphic.geometry;
					else labelGraphic.geometry = graphic.geometry.extent.center;
					// Set the label symbol.
					labelGraphic.symbol = mapObjectUtils.getSymbolForLabel(mapObject);
				}
			}                

			// Add objects to layers.
			if (graphic) layer.add(graphic);
			if (labelGraphic) labelLayer.add(labelGraphic);
		}
	}

	const loadingPromiseState = useRef<{pending: boolean,loading: boolean}>({pending:false,loading:false});

	const updateFootprints = () => {
		if (mapView==null) return;
		if (loadingPromiseState.current.loading===true) { // a load is ongoing?
			loadingPromiseState.current.pending = true; // we want to load again when this completes.
			return;
		}
		loadingPromiseState.current.pending = false; // when we load, we clear the "load pending" flag because we are doing it now.
		// Find the map layers.
		var layer = findLayer(MapObject.TYPE_FOOTPRINT);
		var labelLayer = findLayer(MapObject.TYPE_FOOTPRINT+"-label");
		// Get map extent
		var extent = mapView.extent;
		var targetSpatialReference = new SpatialReference({wkid:4326});
		if (canProject(extent,targetSpatialReference)) {
			extent = project(extent,targetSpatialReference) as Extent;
		}
		var bounds = Bounds.fromXY("4326",extent.xmin,extent.ymin,extent.xmax,extent.ymax);
		// Load footprint data
		viewer.getFootprints2D(bounds,mapView.scale)
			.finally(() => {
				loadingPromiseState.current.loading = false;
				if (loadingPromiseState.current.pending) {
					updateFootprints();
				}
			})
			.then(mapObjects => {
				// Clear objects
				layer.removeAll();
				labelLayer.removeAll();
				// Add objects 
				addMapObjectsToLayer(layer,labelLayer,mapObjects);
				// // Repaint
				// layer.redraw();
				// labelLayer.redraw();
			})
			.catch(reason => {
				console.error("Error calling getFootprints2D(",bounds,")",reason);
			})
	};

	const updateGlobals = () => {
		if (mapView) {
			// Find the map layers.
			var layer = findLayer(MapObject.TYPE_GLOBAL);
			// Clear objects
			layer.removeAll();
			// Add objects 
			let mapObjects = viewer.getGlobals();
			addMapObjectsToLayer(layer,null,mapObjects);
		}
	}

	const updateMeasurements = () => {
		if (mapView) {
			// Find the map layers.
			var layer = findLayer(MapObject.TYPE_MEASUREMENT);
			// Clear objects
			layer.removeAll();
			// Add objects 
			let measureState = viewer.getMeasureState();
			addMapObjectsToLayer(layer,null,measureState.mapObjects);
		}
	}

    /**
	 * Update the viewer field-of-views.
	 */
	const updateViews = () => {
		if (mapView) {
			// Find the map layers
			var layer = findLayer(MapObject.TYPE_FOV);
			var labelLayer = findLayer(MapObject.TYPE_FOV+"-labels");
			// Clear objects
			layer.removeAll();
			labelLayer.removeAll();
			// Get field-of-view objects
			var mapObjects = viewer.getFieldOfViews2D();
			// Add objects
			addMapObjectsToLayer(layer,labelLayer,mapObjects);
		}
	}

	/**
	 * 
	 * @param evt Handle a map move event.
	 */
	const onMapMove = () => {

		updateFootprints();

		// We are not reacting to a change we did?
		if (updatingMap.current==false) {
			const mapPoint: Point = mapView.toMap({x: (mapView.width/2), y: (mapView.height/2)});
			const wgsSpatialRef = new SpatialReference({wkid:4326});

			if (canProject(mapPoint.spatialReference,wgsSpatialRef)) {
				let mapPoint2 = project(mapPoint,wgsSpatialRef) as Point;
				let coordinate = Coordinate.fromXY("4326",mapPoint2.x,mapPoint2.y);
				viewer.setReferenceViewPosition(coordinate);
			}
		}
	}

	/**
	 * Handle a map click event.
	 */
	const onMapClick = (evt) => {
		if (waitingForPickPosition==true) {

			const mapPoint: Point = mapView.toMap({x: evt.x,y: evt.y});		
            const wgsSpatialRef = new SpatialReference({wkid:4326});

            if (canProject(mapPoint.spatialReference,wgsSpatialRef)) {
                let mapPoint2 = project(mapPoint,wgsSpatialRef) as Point;
                let coordinate = Coordinate.fromXY("4326",mapPoint2.x,mapPoint2.y);
				viewer.setPickPosition(coordinate)
            }
	
			// Clear cursor and flag.
			mapView.container.style.cursor = "auto";
			setWaitingForPickPosition(false);
		} 
	};

	/**
	 * Handle a "zoom to extent" signal from the viewer.
	 */
	const handleZoomToExtent = (orbitExtent) => {
		// Create esri Extent
		let extent = new Extent({
			xmin:orbitExtent.minX,
			ymin:orbitExtent.minY,
			xmax:orbitExtent.maxX,
			ymax:orbitExtent.maxY,
			spatialReference:new SpatialReference({wkid:4326})
		});
		if (canProject(extent,mapView)) {
			// Project to map crs
			extent = project(extent,mapView) as Extent;
			// Set extent
			updatingMap.current = true;
			mapView.extent = extent;
			updatingMap.current = false;
		}
	}

	const handlePanToPosition = (position) => {
		// Create esri Extent
		let extent = new Point({
			x:position.x,
			y:position.y,
			spatialReference:new SpatialReference({wkid:4326})
		});
		if (canProject(extent,mapView)) {
			// Project to map crs
			let point = project(extent,mapView) as Point;
			// Set extent

			updatingMap.current = true;
			mapView.center = point;
			updatingMap.current = false;
		}
	}

	/**
	 * Handle a change in the active map view.
	 */
	const activeViewChangeHandler = (jmv: JimuMapView) => {
		if (jmv && jmv.view instanceof MapView) {
			setMapView(jmv.view);
		} else {
			setMapView(null);
		}
	}

	return (
		<div ref={widgetRef} className="widget-demo jimu-widget m-0">
			{
				widgetProps.useMapWidgetIds &&
				widgetProps.useMapWidgetIds.length === 1 && (
					<JimuMapViewComponent
						useMapWidgetId={widgetProps.useMapWidgetIds?.[0]}
						onActiveViewChange={activeViewChangeHandler}
					/>
				)
			}
			<div ref={elementRef} />
		</div>
	)
}

export default WidgetView;